#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import Any

from onetl.base import BaseDBConnection


class SupportWhereNone:
    @classmethod
    def validate_where(
        cls,
        connection: BaseDBConnection,
        where: Any,
    ) -> None:
        if where is not None:
            raise TypeError(f"'where' parameter is not supported by {connection.__class__.__name__}")
