#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from onetl.file.format.avro import Avro
from onetl.file.format.csv import CSV
from onetl.file.format.excel import Excel
from onetl.file.format.json import JSON
from onetl.file.format.jsonline import JSONLine
from onetl.file.format.orc import ORC
from onetl.file.format.parquet import Parquet
from onetl.file.format.xml import XML
