#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import logging
from typing import ClassVar

from onetl.strategy.base_strategy import BaseStrategy
from onetl.strategy.snapshot_strategy import SnapshotStrategy

log = logging.getLogger(__name__)


class StrategyManager:
    default_strategy: ClassVar[type] = SnapshotStrategy

    _stack: ClassVar[list[BaseStrategy]] = []

    @classmethod
    def push(cls, strategy: BaseStrategy) -> None:
        cls._stack.append(strategy)

    @classmethod
    def pop(cls) -> BaseStrategy:
        return cls._stack.pop()

    @classmethod
    def get_current_level(cls) -> int:
        return len(cls._stack)

    @classmethod
    def get_current(cls) -> BaseStrategy:
        if cls._stack:
            return cls._stack[-1]

        return cls.default_strategy()
